{-# LANGUAGE BangPatterns #-}
{-# LANGUAGE MagicHash #-}
{-# LANGUAGE ScopedTypeVariables #-}
module Engine.Eval
( EvalResult(..)
, CompiledExpression(..)
, evalWithSeq
) where

import Control.Exception
import CoreLint
import CoreSyn
import CoreUtils (exprType)
import Data.List
import Engine.EngineHValue
import FastString
import GHC hiding (compileExpr, exprType)
import Outputable
import GHC.Prim
import HscMain (compileExpr)
import HscTypes
import Id
import MkCore
import Prelude hiding (catch)
import Util.GHC
import TysWiredIn (unitDataConId)
import UniqSupply
import Control.Applicative

data EvalResult 
  = EvalOk !EngineHValue
  | EvalException !SomeException

data CompiledExpression = CE HValue !Type
  
evalWithSeq :: CompiledExpression -> IO EvalResult
evalWithSeq (CE hval etipe) = do
  (hval `seq` do
    ehv <- liftIO $ (mkEngineHValue hval etipe)
    return $ EvalOk ehv
   ) `catch` \(e :: SomeException) -> do
    e' <- makeSafe e
    return (EvalException e')

makeSafeStr :: String -> IO String
makeSafeStr str = do
  (case str of
    ((!x):xs) -> do
      rest <- makeSafeStr xs
      rest `seq` return (x:rest)
    []     -> return "") `catch` (\(_ :: SomeException) -> return "<nested exception thrown>")
    
makeSafe :: SomeException -> IO SomeException
makeSafe e = (do
  () <- return (foldl' seq '!' str `seq` ())
  return e) `catch` (\(_ :: SomeException) -> do
    str' <- makeSafeStr str
    return (SomeException (ErrorCall str'))
    )
  where
    str = show e
