{-# LANGUAGE BangPatterns #-}
{-# LANGUAGE TemplateHaskell #-}
module FrontEnd.Options where

import Data.Int
import Distribution.Simple.Setup
import Plan (Depth)
import Engine (Strategy, fifoStrategy)
import IdDB (IdSource(..))
import Util.FCLabels
import Data.Monoid

data RuntimePlan
  = DFS
  | Random
  | RestartingRandom
  | DFSEvenTime
  deriving Eq

data IrulanOptions
  = IrulanOptions { runtimePlanType        :: !(Flag RuntimePlan)
                  , maximumDepth           :: !(Flag Depth)  -- { for depth bounded plans }
                  , maximumExpressionSize  :: !(Last (Maybe Int))    -- { for depth bounded plans }
                  , maximumExpressionDepth :: !(Last (Maybe Int))    -- { for depth bounded plans }
                  , maximumRuntime         :: !(Flag Double) -- seconds  { for Tree and random-like plans }
                  , maxExpressionTime      :: !(Flag Int)    -- microseconds (1/10^6)
                  , maxExpressionAllocs    :: !(Flag Int64)  -- bytes
                  , outputExperimentTrace  :: !(Flag Bool)
                  , outputDebugInformation :: !(Flag Bool)
                  , outputStatistics       :: !(Flag FilePath)
                  , useSimpleOutput        :: !(Flag Bool)
                  , useCaseStatement       :: !(Flag Bool)
                  , useCaseArgument        :: !(Flag Bool)
                  , sourceIdsFrom          :: !(Flag IdSource)
                  , explicitIdSource       :: !([String])
                  , cache                  :: !(Last (Maybe Strategy))
                  , quiet                  :: !(Flag Bool)
                  , onlyTestProperties     :: !(Flag Bool)
                  , doProfiling            :: !(Flag Bool)
                  , intConstants           :: ![Int]
                  , integerConstants       :: ![Integer]
                  , charConstants          :: ![Char]
                  , floatConstants         :: ![Float]
                  , doubleConstants        :: ![Double]
                  , dumpPlan               :: !(Flag Bool)
                  , doIterativeDeepening   :: !(Flag Bool)
                  }

mkLabels [''IrulanOptions]



defaultOptions :: IrulanOptions
defaultOptions
  = IrulanOptions { runtimePlanType        = Flag $ DFS
                  , maximumDepth           = Flag $ 5
                  , maximumExpressionSize  = Last . Just . Just $ 5
                  , maximumExpressionDepth = Last . Just . Just $ 5
                  , maximumRuntime         = Flag $ 10
                  , maxExpressionTime      = Flag $ 1 * (10^ (6 :: Integer)) -- 1 seconds
                  , maxExpressionAllocs    = Flag $ 1024 * 1024 * 128 -- 128Meg
                  , outputExperimentTrace  = Flag $ False
                  , outputDebugInformation = Flag $ False
                  , outputStatistics       = NoFlag
                  , useSimpleOutput        = Flag $ True
                  , useCaseStatement       = Flag $ True
                  , useCaseArgument        = Flag $ False
                  , sourceIdsFrom          = Flag $ SourceModule
                  , explicitIdSource       = []
                  , cache                  = Last . Just $ Just (fifoStrategy 1000)
                  , quiet                  = Flag $ False
                  , onlyTestProperties     = Flag $ False
                  , doProfiling            = Flag $ False
                  , intConstants           = []
                  , integerConstants       = []
                  , charConstants          = []
                  , floatConstants         = []
                  , doubleConstants        = []
                  , dumpPlan               = Flag $ False
                  , doIterativeDeepening   = Flag $ False
                  }


instance Monoid IrulanOptions where
  mempty 
    = IrulanOptions { runtimePlanType        = mempty
                    , maximumDepth           = mempty
                    , maximumExpressionSize  = mempty
                    , maximumExpressionDepth = mempty
                    , maximumRuntime         = mempty
                    , maxExpressionTime      = mempty
                    , maxExpressionAllocs    = mempty
                    , outputExperimentTrace  = mempty
                    , outputDebugInformation = mempty
                    , outputStatistics       = mempty
                    , useSimpleOutput        = mempty
                    , useCaseStatement       = mempty
                    , useCaseArgument        = mempty
                    , sourceIdsFrom          = mempty
                    , explicitIdSource       = mempty
                    , cache                  = mempty
                    , quiet                  = mempty
                    , onlyTestProperties     = mempty
                    , doProfiling            = mempty
                    , intConstants           = mempty
                    , integerConstants       = mempty
                    , charConstants          = mempty
                    , floatConstants         = mempty
                    , doubleConstants        = mempty
                    , dumpPlan               = mempty
                    , doIterativeDeepening   = mempty
                    }

  mappend a b 
    = IrulanOptions { runtimePlanType        = combine runtimePlanType
                    , maximumDepth           = combine maximumDepth
                    , maximumExpressionSize  = combine maximumExpressionSize
                    , maximumExpressionDepth = combine maximumExpressionDepth
                    , maximumRuntime         = combine maximumRuntime       
                    , maxExpressionTime      = combine maxExpressionTime
                    , maxExpressionAllocs    = combine maxExpressionAllocs
                    , outputExperimentTrace  = combine outputExperimentTrace
                    , outputDebugInformation = combine outputDebugInformation
                    , outputStatistics       = combine outputStatistics
                    , useSimpleOutput        = combine useSimpleOutput
                    , useCaseStatement       = combine useCaseStatement
                    , useCaseArgument        = combine useCaseArgument
                    , sourceIdsFrom          = combine sourceIdsFrom
                    , explicitIdSource       = combine explicitIdSource
                    , cache                  = combine cache
                    , quiet                  = combine quiet
                    , onlyTestProperties     = combine onlyTestProperties
                    , doProfiling            = combine doProfiling
                    , intConstants           = combine intConstants
                    , integerConstants       = combine integerConstants
                    , charConstants          = combine charConstants
                    , floatConstants         = combine floatConstants
                    , doubleConstants        = combine doubleConstants
                    , dumpPlan               = combine dumpPlan
                    , doIterativeDeepening   = combine doIterativeDeepening
                    }
      where
        combine f = f a `mappend` f b
