{-# LANGUAGE ViewPatterns #-}
{-# LANGUAGE PatternGuards #-}
{-# LANGUAGE BangPatterns #-}
module IdDB.TypeMap
( lookupProviders
, addProvider
, addDCProvider
, addIntProvider
, addAtType
, allIds
, empty
, TypeMap
, TypeMapDC
, TypeMapInt
, AbsTypeMap(..)
, TypeMapNode(..)
, TypeMapPath(..)
, containsTyCon
) where

import Control.Arrow
import DataCon
import Data.List
import Data.Monoid
import Name
import OccName
import Outputable (showPpr)
import qualified Util.UniqMap as UniqMap
import qualified Util.UniqSet as UniqSet
import TyCon
import Type
import Util.GHC
import Util (trace)
import Util.UniqMap (UniqMap)
import Util.UniqSet (UniqSet)
import Var
import VarEnv

data TypeMapNode a
  = Empty {- An optimisation case -}
  | Split { tyVars    :: !(UniqMap TyVar a)
          , appTys    :: !(TypeMapNode (TypeMapNode a))
          , tyCons    :: !(UniqMap TyCon (TypeMapPath a))
          , funTys    :: !(TypeMapNode (TypeMapNode a))
          , forallTys :: !(UniqMap TyVar (TypeMapNode a))
          }


type TypeMap    = AbsTypeMap Id
type TypeMapDC  = AbsTypeMap DataCon
type TypeMapInt = AbsTypeMap Int

data AbsTypeMap a = ATM !(TypeMapNode (UniqSet a)) !InScopeSet

data TypeMapPath a
  = Stop
  | Step { stopsHere :: !a 
         , carrysOn  :: !(TypeMapNode (TypeMapPath a))
         }

empty :: AbsTypeMap a
empty = ATM Empty emptyInScopeSet

lookupProviders :: Type -> AbsTypeMap a -> [(TvSubst, UniqSet a)]
lookupProviders ty (ATM tm iss) = {- trace ("Lookup Providers: " ++ showDebug ty ++ " gives " ++ showDebug res) $ -} res
  where
    ty' = alphaConvertRawTy iss ty
    res = realLookupTypeProviders ty' emptyTvSubst tm 

realLookupTypeProviders :: (Monoid a) => Type -> TvSubst -> TypeMapNode a -> [(TvSubst, a)]

{-
realLookupTypeProviders ty subst node
  | False <- trace ("realLookupTypeProviders: " ++ showDebug ty ++ ", " ++ showDebug subst) True = error "urk"
-}

realLookupTypeProviders _ _ Empty = []

realLookupTypeProviders t@(getTyVar_maybe -> Just _) subst tm
  = do
      (valTy, val) <- flattenTypeMap tm
      let valTy' = substTy subst valTy
      let t' = substTy subst t

      case unifyTypes valTy' t' of
        Nothing -> {- trace ("realLookupTypeProviders: unifyTypes failed: " ++ showDebug valTy' ++ " vs " ++ showDebug t') -} []
        Just subst' -> return (subst' `composeTvSubsts` (subst `unionSubsts` subst'), val)

realLookupTypeProviders t@(splitAppTy_maybe -> Just (t1,t2)) subst tm 
  = lookupInTyVars t subst tm $ do
      (subst', part2) <- realLookupTypeProviders t1 subst (appTys tm)
      realLookupTypeProviders t2 subst' part2

realLookupTypeProviders t@(splitTyConApp_maybe -> Just (tc,tys)) subst tm 
  = lookupInTyVars t subst tm $ 
      pathLookupTypeProviders tys path subst
  where
    path = UniqMap.lookup (tyCons tm) tc

realLookupTypeProviders t@(splitFunTy_maybe -> Just (t1,t2)) subst tm 
  = lookupInTyVars t subst tm $ do
      (subst', part2) <- realLookupTypeProviders t1 subst (funTys tm)
      realLookupTypeProviders t2 subst' part2

realLookupTypeProviders t@(splitForAllTy_maybe -> Just (tv,ty)) subst tm
  = lookupInTyVars t subst tm $
      realLookupTypeProviders ty subst part2 
  where
    part2 = UniqMap.lookup (forallTys tm) tv

lookupInTyVars :: Type -> TvSubst -> TypeMapNode a -> [(TvSubst, a)] -> [(TvSubst, a)]
lookupInTyVars ty subst (Split { tyVars = tyVars }) rest = tyVarOptions ++ rest
  where
    tyVarOptions = do
      let ty' = substTy subst ty
      (tv, rhs) <- UniqMap.elems tyVars
      let rhsTy = substTy subst (mkTyVarTy tv)

      Just subst' <- return $ unifyTypes ty' rhsTy
      
      return (subst' `composeTvSubsts` (subst `unionSubsts` subst'), rhs)

flattenTypeMap :: TypeMapNode a -> [(Type, a)]
flattenTypeMap Empty = []
flattenTypeMap split = tyVars' ++ appTys' ++ tyCons' ++ funTys' ++ forallTys'
  where
    tyVars' = map (first mkTyVarTy) . UniqMap.elems . tyVars $ split
    appTys' = do
      (l, next) <- flattenTypeMap (appTys split)
      (r, a)    <- flattenTypeMap next
      return (mkAppTy l r, a)
    tyCons' = do
      (tc, path) <- UniqMap.elems . tyCons $ split
      (tys, a)   <- flattenTypePath path
      return (mkTyConApp tc tys, a)
    funTys' = do
      (l, next) <- flattenTypeMap (funTys split)
      (r, a)    <- flattenTypeMap next
      return (mkFunTy l r, a)
    forallTys' = do
      (tv, next) <- UniqMap.elems . forallTys $ split
      (r, a)     <- flattenTypeMap next
      return (mkForAllTy tv r, a)
      

flattenTypePath :: TypeMapPath a -> [([Type], a)]
flattenTypePath Stop = []
flattenTypePath step = ([], stopsHere step) : do
  (ty, tp) <- flattenTypeMap . carrysOn $ step
  (tys, a) <- flattenTypePath tp
  return (ty:tys, a)

allIds :: AbsTypeMap a -> UniqSet a
allIds (ATM mp _) = mconcat . map snd . flattenTypeMap $ mp

pathLookupTypeProviders :: Monoid a => [Type] -> TypeMapPath a -> TvSubst -> [(TvSubst, a)]
pathLookupTypeProviders _ Stop _ = []
pathLookupTypeProviders [] step subst = [(subst, stopsHere step)]
pathLookupTypeProviders (t:ts) step subst = do
  (subst', path') <- realLookupTypeProviders t subst (carrysOn step)
  pathLookupTypeProviders ts path' subst'
  

{- Adds all the types Id can provide -}
addProvider :: TypeMap -> Id -> TypeMap
addProvider itm@(ATM tm env) theId 
  | "unsafe" `isPrefixOf` (occNameString . getOccName $ theId) = itm
  | otherwise = ATM (foldr (\ty -> addAtType ty singleSet) tm suppliedTypes) env'
  where
    (env', theId', _) = alphaConvert env theId
    singleSet = UniqSet.singleton theId
    idType = varType theId'
    (_, rawType) = splitForAllTys idType
    
    rawTypeParts :: [Type]
    (rawTypeParts, lastPart) = splitFunTys rawType

    suppliedTypes :: [Type]
    suppliedTypes = map (`mkFunTys` lastPart) (tails rawTypeParts)

addDCProvider :: TypeMapDC -> DataCon -> TypeMapDC
addDCProvider (ATM tm env) theDC
  = ATM (foldr (\ty -> addAtType ty singleSet) tm rawTypeParts) env'
    where
      (env', theId', _) = alphaConvert env (dataConWrapId theDC)
      singleSet = UniqSet.singleton theDC


      idType = varType theId'
      (_, rawType) = splitForAllTys idType
      (rawTypeParts, _) = splitFunTys rawType

addIntProvider :: TypeMapInt -> Type -> Int -> TypeMapInt
addIntProvider (ATM tm env) ty theInt
  = ATM (addAtType tyToAdd (UniqSet.singleton theInt) tm) env'
  where
    (env', ty', _) = alphaConvertTy env ty
    tyToAdd = dropForAlls ty'


addAtType :: Monoid a => Type -> a -> TypeMapNode a -> TypeMapNode a
addAtType ty it Empty = addAtType ty it emptySplit
addAtType (getTyVar_maybe -> Just tv)  it s = s { tyVars = UniqMap.insert (tyVars s) tv it }
addAtType (splitAppTy_maybe -> Just (t1,t2)) it s = s { appTys = addAtType t1 part2 (appTys s) }
  where
    part2 = addAtType t2 it mempty
addAtType (splitTyConApp_maybe -> Just (tc,tys)) it s = s { tyCons = UniqMap.insert (tyCons s) tc part2 }
  where
    part2 = buildPath tys it
addAtType (splitFunTy_maybe -> Just (t1,t2)) it s = s { funTys = addAtType t1 part2 (funTys s) }
  where
    part2 = addAtType t2 it mempty
addAtType (splitForAllTy_maybe -> Just (tv,ty)) it s = s { forallTys = UniqMap.insert (forallTys s) tv part2 }
  where
    part2 = addAtType ty it mempty

buildPath :: Monoid a => [Type] -> a -> (TypeMapPath a)
buildPath [] a     = emptyStep { stopsHere = a }
buildPath (t:ts) a = emptyStep { carrysOn = addAtType t (buildPath ts a) mempty }

instance Monoid a => Monoid (TypeMapPath a) where
  mempty = Stop
  Stop `mappend` s = s
  s `mappend` Stop = s
  s1 `mappend` s2 = Step { stopsHere = stopsHere s1 `mappend` stopsHere s2
                         , carrysOn  = carrysOn s1  `mappend` carrysOn s2
                         }
emptyStep :: Monoid a => TypeMapPath a
emptyStep = Step { stopsHere = mempty
                 , carrysOn  = mempty
                 }

emptySplit :: Monoid a => TypeMapNode a
emptySplit = Split { tyVars    = mempty
                   , appTys    = mempty
                   , tyCons    = mempty
                   , funTys    = mempty
                   , forallTys = mempty
                   }

instance Monoid a => Monoid (TypeMapNode a) where
  mempty = Empty 
  Empty `mappend` s = s
  s `mappend` Empty = s
  s1 `mappend` s2 = Split { tyVars    = tyVars s1 `mappend` tyVars s2
                          , appTys    = appTys s1 `mappend` appTys s2
                          , tyCons    = tyCons s1 `mappend` tyCons s2
                          , funTys    = funTys s1 `mappend` funTys s2
                          , forallTys = forallTys s1 `mappend` forallTys s2
                          }

containsTyCon :: AbsTypeMap a -> TyCon -> Bool
containsTyCon (ATM tm _) ty = realContainsTyCon ty tm
  where
    realContainsTyCon :: TyCon -> TypeMapNode a -> Bool
    realContainsTyCon _ Empty  = False
    realContainsTyCon tc split = UniqMap.member tc (tyCons split)
