{-# LANGUAGE BangPatterns #-}
module Util.UniqMap
( UniqMap
, insert
, member
, singleton
, values
, Util.UniqMap.null
, Util.UniqMap.lookup
, Util.UniqMap.elems
, Util.UniqMap.delete
, Util.UniqMap.adjust
) where

import UniqFM
import UniqSet
import Data.Monoid
import Unique
import Data.Maybe
import Outputable

data UniqMap key elt = UM !(UniqFM elt) !(UniqSet key)

instance Monoid elt => Monoid (UniqMap key elt) where
  mempty                        = UM emptyUFM emptyUFM
  (UM ufm1 us1) `mappend` (UM ufm2 us2) = UM (plusUFM_C mappend ufm1 ufm2)
                                             (unionUniqSets us1 us2)

instance Functor (UniqMap key) where
  fmap f (UM elts keys) = UM (mapUFM f elts) keys

insert :: (Uniquable key, Monoid elt) => UniqMap key elt -> key -> elt -> UniqMap key elt
insert (UM map keySet) key elt = UM (addToUFM_C mappend map key elt)
                                    (addOneToUniqSet keySet key)

lookup :: (Uniquable key, Monoid elt) => UniqMap key elt -> key -> elt
lookup (UM uMap _) = fromMaybe mempty . lookupUFM uMap

singleton :: (Uniquable key, Monoid elt) => key -> elt -> UniqMap key elt
singleton key elt = insert mempty key elt

elems :: Uniquable key => UniqMap key elt -> [(key, elt)]
elems (UM uMap keySet) = map (\key -> (key, fromMaybe (error "UniqMap.elems key set out of sync with elem map") (lookupUFM uMap key))) (uniqSetToList keySet)

values :: UniqMap key elt -> [elt]
values (UM ufm _) = eltsUFM ufm

member :: Uniquable key => key -> UniqMap key elt -> Bool
member key (UM uMap _) = elemUFM key uMap

adjust :: Uniquable key => (elt -> elt) -> key -> UniqMap key elt -> UniqMap key elt
adjust f key m@(UM uMap uSet) = case lookupUFM uMap key of
                                  Nothing -> m
                                  Just v -> UM (addToUFM uMap key (f v)) (uSet)

null :: UniqMap key elt -> Bool
null (UM m _) = isNullUFM m

delete :: (Uniquable key) => UniqMap key elt -> key -> UniqMap key elt
delete (UM uMap keySet) key = UM (delFromUFM uMap key) (delOneFromUniqSet keySet key)

instance (Outputable elt) => Outputable (UniqMap key elt) where
  ppr (UM m _) = ppr m
