{-# LANGUAGE MultiParamTypeClasses #-}
{-# LANGUAGE FlexibleInstances #-}
{-# LANGUAGE NamedFieldPuns #-}
module Vis.Graph where

import Data.View
import GraphRewriting.Graph
import GraphRewriting.Graph.Write

import Plan
import Engine


data PlanNode = Root { out :: Port
                     }
              | RanExperiment { inp      :: Port
                              , exp      :: Experiment
                              , result   :: ExperimentResult
                              , out      :: Port
                              }
              | Split { inp :: Port
                      , children :: [Port]
                      }
              | PlanToDo { inp :: Port
                         , plan :: TreePlan
                         }
              | Nil { inp :: Port }

instance View [Port] PlanNode where
  inspect Root { out }               = [out]
  inspect RanExperiment { inp, out } = [inp, out]
  inspect Split { inp, children }    = inp:children
  inspect PlanToDo { inp }           = [inp]
  inspect Nil { inp }                = [inp]

  update ports ski = case ski of
    Root {}          -> ski { out = out }               where [out]      = ports
    RanExperiment {} -> ski { inp = inp, out = out }    where [inp, out] = ports
    Split {}         -> ski { inp = inp, children = c } where (inp:c)    = ports
    PlanToDo {}      -> ski { inp = inp }               where [inp]      = ports
    Nil {}           -> ski { inp = inp }               where [inp]      = ports

fromPlan :: TreePlan -> Graph PlanNode
fromPlan tp = flip execGraph emptyGraph $ do
  e1 <- newEdge
  newNode PlanToDo { inp = e1 , plan = tp }
  newNode Root { out = e1 }

