{-# LANGUAGE ViewPatterns #-}
{-# LANGUAGE PatternGuards #-}
{-# LANGUAGE MultiParamTypeClasses #-}
{-# LANGUAGE FlexibleInstances #-}
module AIPlayer(AIPlayer(..)) where

import Player
import Board
import Control.Monad
import Data.List
import Data.Monoid
import Min
import Max
import qualified Data.Foldable as F
import qualified Data.Map as Map
import qualified Data.Set as Set

import Data.Map (Map)
import Data.Set (Set)
import Control.Arrow
import Data.Maybe
import Data.Ord

data AIPlayer = AIPlayer

instance (Monad m) => Player m AIPlayer where
  getMove b aip = return (minimax b, aip)
  
data WLD = Lose | Draw | Win
  deriving (Eq, Show, Bounded, Ord)

minimax :: Board -> Position
minimax = fst . minimaxPlayer

minimaxPlayer :: Board -> (Position, GameEnd)
minimaxPlayer board 
  | (Right eg) <- status = (error "No position!", eg)
  | (Left us)  <- status = bestOutcome us
  where
    status = nextMove board

    nextBoards :: [(Position,Board)]
    nextBoards = F.foldMap ((:[]) . (id &&& fromMaybe (error "nextBoards") . placePiece board))
               . filterSymmetry
               . availablePositions $ board

    bestOutcome :: Piece -> (Position, GameEnd)
    bestOutcome us = foldr1 findMax
                   . map (second $ snd . minimaxPlayer)
                   $ nextBoards
      where
        findMax :: (Position, GameEnd) -> (Position, GameEnd) -> (Position, GameEnd)
        findMax a@(_, ourStatus us -> Win)   _                          = a
        findMax   (_, ourStatus us -> Lose)  a                          = a
        findMax   (_, ourStatus us -> Draw)  a@(_, ourStatus us -> Win) = a
        findMax a                            _                          = a

ourStatus :: Piece -> GameEnd -> WLD
ourStatus us (GameWon winner) 
  | winner == us      = Win
  | otherwise         = Lose
ourStatus _ GameDrawn = Draw

filterSymmetry :: Set Position -> Set Position
filterSymmetry availables
  | availables == allPositions = key3
  | Set.size availables == 8   = symPos8
  | otherwise                  = availables
  where
    key3 = Set.fromList [TL, TM, MM]

    [takenPos] = Set.toList $ allPositions Set.\\ availables 

    syms8 = [(TL, [TM, MM, TR, MR, BR])
            ,(TM, [TL, ML, MM, BL, BM])
            ,(MM, [TL, TM])
            ]

    rotatedTaken :: [Position]
    rotatedTaken = iterate rotate90 $ takenPos

    Just numRotates = findIndex (`Set.member` key3) rotatedTaken
    numRotatesBack  = 4 - numRotates

    Just syms = lookup (rotatedTaken !! numRotates) syms8
    symPos8 = Set.fromList . map ((!! numRotatesBack) . iterate rotate90) $ syms
    


rotate90 :: Position -> Position
rotate90 TL = TR
rotate90 TM = MR
rotate90 TR = BR
rotate90 ML = TM
rotate90 MM = MM
rotate90 MR = BM
rotate90 BL = TL
rotate90 BM = ML
rotate90 BR = BL

