module Console where
import Player
import AIPlayer
import Board
import Control.Applicative
import ConsolePlayer
import RandomPlayer
import System.IO.UTF8 as UTF8

data ConsoleGame
  = CG
  { board :: Board
  , noughtsPlayer :: EPlayer IO
  , crossesPlayer :: EPlayer IO
  }

main :: IO ()
main = do
  rp <- reallyRandomPlayer
  let aip = AIPlayer
  unfoldM playStep $ CG { board         = emptyBoard
                        , noughtsPlayer = EPlayer aip
                        , crossesPlayer = EPlayer rp
                        }

playStep :: ConsoleGame -> IO (Maybe ConsoleGame)
playStep cg@(CG { board = board
               , noughtsPlayer = pn
               , crossesPlayer = pc
               }) = do
  UTF8.putStr formattedBoard
  case nm of
    (Left Nought) -> Just <$> playWith pn (\pn' -> cg { noughtsPlayer = pn' })
    (Left Cross)  -> Just <$> playWith pc (\pc' -> cg { crossesPlayer = pc' })
    (Right _)     -> endGame
  where
    nm = nextMove board

    playWith :: EPlayer IO -> (EPlayer IO -> ConsoleGame) -> IO ConsoleGame
    playWith player update = do
      (pos, player') <- getMove board player
      let Just board' = placePiece board pos
      return $ (update player') { board = board' }

    endGame :: IO (Maybe ConsoleGame)
    endGame = return Nothing

    formattedBoard
      = unlines [ ""
                , "┌" ++ "─"  ++ "┬" ++ "─"  ++ "┬" ++ "─"  ++ "┐"
                , "│" ++ p TL ++ "│" ++ p TM ++ "│" ++ p TR ++ "│"
                , "├" ++ "─"  ++ "┼" ++ "─"  ++ "┼" ++ "─"  ++ "┤"
                , "│" ++ p ML ++ "│" ++ p MM ++ "│" ++ p MR ++ "│"
                , "├" ++ "─"  ++ "┼" ++ "─"  ++ "┼" ++ "─"  ++ "┤"
                , "│" ++ p BL ++ "│" ++ p BM ++ "│" ++ p BR ++ "│"
                , "└" ++ "─"  ++ "┴" ++ "─"  ++ "┴" ++ "─"  ++ "┘"
                , ""
                , bottomLine
                ]
      where
        p :: Position -> String
        p pos = maybe " " formatPiece (lookupPiece pos board)
        bottomLine = case nm of
          Left piece            -> formatPiece piece ++ " to move:"
          Right GameDrawn       -> "Game Drawn!"
          Right (GameWon piece) -> "Congratulations, " ++ formatPiece piece ++ " won!"

        formatPiece :: Piece -> String
        formatPiece Nought = "○"
        formatPiece Cross  = "✗"

unfoldM :: (Monad m) => (a -> m (Maybe a)) -> a -> m ()
unfoldM f init = do
  res <- f init
  case res of
    Nothing     -> return ()
    (Just next) -> unfoldM f next
