{-# LANGUAGE TemplateHaskell #-}
module IdDB.ConstructorGraphTest
( tests
) where

import IdDB.ConstructorGraphTestUtil
import IdDB.ConstructorGraphTestLib
import Test.HUnit ((~:))

unit             = t 0 [t|()|]
maybeUnit        = t 1 [t|Maybe ()|]
listUnit         = t 2 [t|[()]|]
divergingUnit    = t 3 [t|Diverging ()|]
tuple2Unit       = t 4 [t|((),())|]
tupleUnitInt     = t 5 [t|((),Int)|]
int              = t 6 [t|Int|]
maybeInt         = t 7 [t|Maybe Int|]
listInt          = t 8 [t|[Int]|]
listTupleUnitInt = t 9 [t|[((),Int)]|]

diverging0Unit = divergingUnit
diverging1Unit = t 10 [t|Diverging (Diverging ())|]
diverging2Unit = t 11 [t|Diverging (Diverging (Diverging ()))|]
diverging3Unit = t 12 [t|Diverging (Diverging (Diverging (Diverging ())))|]
diverging4Unit = t 13 [t|Diverging (Diverging (Diverging (Diverging (Diverging ()))))|]


test 0 = with [] [ unit `gives` []
                 ]

test 1 = with [ maybeUnit ] [ unit      `gives` [ maybeUnit ==> [ ('Just, 0) ~~> n ] ]
                            , maybeUnit `gives` []
                            ]

test 2 = with [ listUnit ]  [ unit      `gives` [ listUnit ==> [ ('(:), 0) ~~> n 
                                                               , ('(:), 1) ~~> [ ('(:), 0) ~~> n
                                                                               , ('(:), 1) ~~> [ ('(:), 0) ~~> n
                                                                                               , ('(:), 1) ~~> [ ('(:), 0) ~~> n
                                                                                                               , ('(:), 1) ~~> (!)
                                                                                                               ]
                                                                                               ]
                                                                               ]
                                                               ]
                                                ]
                            , listUnit  `gives` [ listUnit ==> [ ('(:), 1) ~~> [ ('(:), 1) ~~> [ ('(:), 1) ~~> [ ('(:), 1) ~~> (!) ]]]]]
                            , maybeUnit `gives` []
                            ]

test 3 = with_limit 3 [ divergingUnit ] [ diverging0Unit `gives` []
                                        , diverging1Unit `gives` [ divergingUnit ==> [ ('Diverging, 0) ~~> n ]]
                                        , diverging2Unit `gives` [ divergingUnit ==> [ ('Diverging, 0) ~~> [ ('Diverging, 0) ~~> n ]]]
                                        , diverging3Unit `gives` [ divergingUnit ==> [ ('Diverging, 0) ~~> [ ('Diverging, 0) ~~> [ ('Diverging,0) ~~> n ]]]]
                                        , diverging4Unit `gives` []
                                        ]

test 4 = with [ tuple2Unit ] [ unit `gives` [ tuple2Unit ==> [ ('(,), 0) ~~> n
                                                             , ('(,), 1) ~~> n
                                                             ]
                                            ]
                             , tuple2Unit `gives` []
                             ]

test 5 = with [ tupleUnitInt ] [ unit `gives` [ tupleUnitInt ==> [ ('(,), 0) ~~> n
                                                                 ]
                                              ]
                               , int `gives`  [ tupleUnitInt ==> [ ('(,), 1) ~~> n
                                                                 ]
                                              ]
                              , tupleUnitInt `gives` []
                              ]
test 6 = with [ maybeUnit, listUnit ] [ unit `gives` [ maybeUnit ==> [ ('Just, 0) ~~> n ]
                                                     , listUnit  ==> [ ('(:), 0) ~~> n
                                                                     , ('(:), 1) ~~> [ ('(:), 0) ~~> n
                                                                                     , ('(:), 1) ~~> [ ('(:), 0) ~~> n
                                                                                                     , ('(:), 1) ~~> (!)
                                                                                                     ]
                                                                                     ]
                                                                     ]
                                                    ]
                                      ]

test 7 = with [ maybeInt, maybeUnit ] [ unit `gives` [ maybeUnit ==> [ ('Just, 0) ~~> n ] ]
                                      , int  `gives` [ maybeInt  ==> [ ('Just, 0) ~~> n ] ]
                                      ]

test 8 = with [ listTupleUnitInt ] [ unit `gives`    [ listTupleUnitInt ==> [ ('(:), 0) ~~> tupleLeft
                                                                            , ('(:), 1) ~~> [ ('(:), 0) ~~> tupleLeft
                                                                                            , ('(:), 1) ~~> [ ('(:), 0) ~~> tupleLeft
                                                                                                            , ('(:), 1) ~~> (!)
                                                                                                            ]
                                                                                            ]
                                                                            ]
                                                    ]
                                   ]
  where tupleLeft =  [ ('(,), 0) ~~> n ]

test _ = notWrittenYet


tests = "ConstructorGraphTest" ~: buildTests test
