{-# LANGUAGE ScopedTypeVariables #-}
{-# LANGUAGE ViewPatterns #-}
{-# LANGUAGE GeneralizedNewtypeDeriving #-}
{-# LANGUAGE ScopedTypeVariables #-}
module IdDB.IdDBTestLib
( with
, is
, isn't
, there
, ExpectedProvider ((:::))
, (-->)
, notWrittenYet
, buildTests
) where

import Control.Arrow
import Control.Monad
import Data.List
import Data.Maybe
import Data.Monoid
import Data.Ord
import GHC
import GHC.Paths
import IdDB.Impl
import IdDB.TestLib
import MonadUtils
import qualified IdDB.TypeMap as TM
import qualified Test.HUnit as HUnit
import Test.HUnit ((~:), (@?))
import Type
import Util.TH

buildTests :: (Int -> Maybe Test) -> HUnit.Test
buildTests f = HUnit.TestList $ map mkTest . catMaybes . takeWhile isJust . map (f$) $ [0..]

mkTest :: Test -> HUnit.Test
mkTest test = show (testNames test) ~: runGhc (Just libdir) $ do
  dflags <- getSessionDynFlags 
  setSessionDynFlags dflags
  defaultCleanupHandler dflags $ do
    guessTarget ("IdDB.IdDBSampleModule") Nothing >>= \target -> setTargets [target]
    result <- load LoadAllTargets
    liftIO $ (succeeded result) @? "Couldn't load sample module"

    ids <- mapM idFromTHName (testNames test) 
    idDB <- mkIdDBForIds ids (idSourceToUse test) [] mempty

    forM_ (testConstraints test) (checkConstraint idDB)

checkConstraint :: IdDB -> Constraint -> Ghc ()
checkConstraint db (IsNot tyq) = do
  ty <- typeFromTHTypeQ tyq
  let found = rawIdsForType (dropForAlls ty) db
  liftIO $ (null found) @? "Looking up " ++ show ty ++ ", expected: [], got: " ++ show found
checkConstraint db (Is tyq thExps) = do
  ty <- typeFromTHTypeQ tyq
  exps :: [(TvSubst -> IO (), Id)] <- mapM mkExp thExps

  let found :: [(TvSubst, Id)] = map (snd &&& fst) $ rawIdsForType (dropForAlls ty) db
  let srt = sortBy (comparing snd)

  liftIO $ (length found == length exps) @? "Didn't get equal numbers of expected and found!  Expected: " ++ show (map snd exps) ++ ", got: " ++ show (map snd found)
  liftIO $ zipWithM_ checkOne (srt exps) (srt found)

