{-# LANGUAGE ScopedTypeVariables #-}
{-# LANGUAGE TemplateHaskell #-}
module IdDB.TypeMapTest where

import IdDB.TypeMapTestLib
import IdDB.TypeMapSampleModule
import Test.HUnit ((~:))

bool      = [t|Bool|]
int       = [t|Int|]
char      = [t|Char|]
maybeBool = [t|Maybe Bool|]
maybeInt  = [t|Maybe Int|]


test 0 = with ['true] $ do
          bool `is` 'true
          int  `isn't` there
          
test 1 = with ['true, 'false] $ do
          bool `is` ['true, 'false]
          int `isn't` there

test 2 = with ['true, 'false, 'three] $ do
          bool `is` ['true, 'false]
          int `is`  'three
          char `isn't` there

test 3 = with ['maybeTrue, 'true] $ do
          bool `is` 'true
          maybeBool `is` 'maybeTrue
          int `isn't` there

test 4 = with ['myId] $ do
          bool `is` ['myId ::: (bool --> bool)]

test 5 = with ['intBoolCharMaybeIntMaybeBool, 'true, 'maybeTrue] $ do
          bool `is` 'true
          maybeBool `is` ['intBoolCharMaybeIntMaybeBool, 'maybeTrue]
          (maybeInt --> maybeBool) `is` 'intBoolCharMaybeIntMaybeBool
          char `isn't` there

test 6 = with ['true, 'maybeFalse, 'maybeAny] $ do
          bool `is` 'true
          maybeBool `is` ['maybeFalse ::: maybeBool
                         ,'maybeAny   ::: maybeBool
                         ]
          [t|forall a . Maybe a|] `is` [ 'maybeFalse ::: maybeBool
                                       , 'maybeAny   ::: [t|forall a . Maybe a|]
                                       ]

test 7 = with ['ib, 'ic, 'maybeC, 'maybeTrue, 'true] $ do
          bool `is` ['true, 'ib]
          maybeBool `is` ['maybeTrue, 'maybeC]
          char `is` 'ic
          [t|forall a . Maybe a|] `is` [ 'maybeTrue ::: maybeBool
                                       , 'maybeC ::: (maybeInt --> maybeBool)
                                       ]

test 8 = with ['Just, 'Nothing] $ do
          [t| forall a . Maybe a |] `is` [ 'Nothing ::: [t|forall a . Maybe a|]
                                         , 'Just    ::: [t|forall a . a -> Maybe a|]
                                         ]

test 9 = with ['unsafeFunction] $ do
          int `isn't` there 

test 10 = with ['ib] $ do
            [t| Int -> Bool |] `is` [ 'ib ]

test _ = notWrittenYet

tests = "TypeMapTest" ~: buildTests test
