{-# LANGUAGE DeriveDataTypeable #-}
module Util.GraphTest where

import Data.Graph.Inductive
import Test.HUnit

import Util.Graph

import Data.Typeable
import Data.List
import Data.List

tests = "GraphTest" ~: [test1, test2, test3, test4]

test1 = empty `equal` actual ~? ("Empty minimized: " ++ show actual)
  where
    actual :: Gr () ()
    actual = minimizeGraph empty


test2 = check "test2" init end
  where
    init = [ (0, True) --| () --> (1, False)
           , (0, True) --| () --> (2, False)
           ]

    end  = [ (0, True) --| () --> (1, False)
           , (0, True) --| () --> (1, False)
           ]

test3 = check "test3" init end
  where
    init = [ (0, True) --| () --> (1, False)
           , (2, True) --| () --> (3, True)
           ]

    end = init

test4 = check "test4" init end
  where
    init = [ (0, "S") --| "x" --> (2, "A")
           , (0, "S") --| "y" --> (3, "B")
           , (0, "S") --| "." --> (5, "X")
           , (1, "S") --| "x" --> (2, "A")
           , (1, "S") --| "y" --> (3, "B")
           , (1, "S") --| "." --> (6, "X")
           , (2, "A") --| "a" --> (4, "C")
           , (3, "B") --| "b" --> (4, "C")
           ]

    end  = [ (0, "S") --| "x" --> (2, "A")
           , (0, "S") --| "y" --> (3, "B")
           , (0, "S") --| "." --> (5, "X")
           , (2, "A") --| "a" --> (4, "C")
           , (3, "B") --| "b" --> (4, "C")
           ]

check :: (Ord a, Ord b, Show a, Show b) => String -> [TestEdge a b] -> [TestEdge a b] -> Test
check msg init end = expectEnd `equal` actualEnd ~? ("*" ++ msg ++ "*\n Got: " ++ show actualEnd ++ "\nExpected: " ++ show expectEnd) 
  where
    realInit  = build init
    expectEnd = build end
    actualEnd = minimizeGraph realInit

build :: (Eq a, Ord a) => [TestEdge a b] -> Gr a b
build tedges = insEdges edges
             . insNodes nodes
             $ empty
  where
    nodes  = nub . concatMap (\(TE l _ r) -> [l,r]) $ tedges
    edges  = map (\(TE l e r) -> (fst l, fst r, e)) $ tedges

infixr 9 -->
infixr 0 --|

(--|) :: LNode a -> TestRHS a b -> TestEdge a b
lnod --| (RHS edg rnod) = TE lnod edg rnod

(-->) :: b -> LNode a -> TestRHS a b
edg --> rnod = RHS edg rnod

data TestEdge a b = TE (LNode a) b (LNode a)
data TestRHS a b = RHS b (LNode a)
